use strict;
use warnings;

# this test was generated with Dist::Zilla::Plugin::Test::EOL 0.19

use Test::More 0.88;
use Test::EOL;

my @files = (
    'bin/update_sanctions_csv',
    'lib/Data/Validate/Sanctions.pm',
    'lib/Data/Validate/Sanctions/Fetcher.pm',
    'lib/Data/Validate/Sanctions/Redis.pm',
    't/00-check-deps.t',
    't/00-compile.t',
    't/00-report-prereqs.dd',
    't/00-report-prereqs.t',
    't/01_basic.t',
    't/02_env.t',
    't/03_oo.t',
    't/04_fetcher.t',
    't/05_sanctions_redis.t',
    't/06_fetcher_sources.t',
    't/data/sample_eu.xml',
    't/data/sample_hmt.csv',
    't/data/sample_moha.xml',
    't/data/sample_ofac_consolidated.xml',
    't/data/sample_unsc.xml',
    't/data/sdn.xml',
    't/rc/perlcriticrc',
    't/rc/perltidyrc',
    'xt/20_update.t',
    'xt/author/critic.t',
    'xt/author/distmeta.t',
    'xt/author/eol.t',
    'xt/author/minimum-version.t',
    'xt/author/mojibake.t',
    'xt/author/no-tabs.t',
    'xt/author/pod-syntax.t',
    'xt/author/portability.t',
    'xt/author/test-version.t',
    'xt/release/common_spelling.t',
    'xt/release/cpan-changes.t'
);

eol_unix_ok($_, { trailing_whitespace => 1 }) foreach @files;
done_testing;
